##
## Constant-time SSSE3 AES core implementation.
## version 0.1
##
## By Mike Hamburg (Stanford University), 2009
## Public domain.
##

## load higher-level routines
.include "aes.s"
	
	.text
##
##  _aes_preheat
##
##  Fills register %r10 -> .aes_consts (so you can -fPIC)
##  and %xmm8-%xmm15 as specified below.
##
_aes_preheat:
	lea	.aes_consts(%rip), %r10
	movdqa	         (%r10), %xmm9  # 0F
	movdqa	.k_inv   (%r10), %xmm10 # inv
	movdqa	.k_inv+16(%r10), %xmm11 # inva
	movdqa	.k_sb1   (%r10), %xmm13 # sb1u
	movdqa	.k_sb1+16(%r10), %xmm12 # sb1t
	movdqa	.k_sb2   (%r10), %xmm15 # sb2u
	movdqa	.k_sb2+16(%r10), %xmm14 # sb2t
	ret

##
##  _aes_encrypt_core
##
##  AES-encrypt %xmm0.
##
##  Inputs:
##     %xmm0 = input
##     %xmm9-%xmm15 as in .aes_preheat
##     %r10  = .aes_consts
##    (%rdx) = scheduled keys
##
##  Output in %xmm0
##  Clobbers  %xmm1-%xmm4, %r9, %r11, %rax
##  Preserves %xmm5 - %xmm8 so you get some local vectors
##
##
	.align 16
_aes_encrypt_core:
	mov	%rdx,	%r9
	mov	$16,	%r11
	mov	(%rdx),	%rax
	movdqa	.k_ipt   (%r10), %xmm2 # iptlo
	movdqa	%xmm9,	%xmm1
	pandn	%xmm0,	%xmm1
	psrld	$4,	%xmm1
	pand	%xmm9,	%xmm0
	pshufb	%xmm0,	%xmm2
	movdqa	.k_ipt+16(%r10), %xmm0 # ipthi
	pshufb	%xmm1,	%xmm0
	pxor	16(%r9),%xmm2
	pxor	%xmm2,	%xmm0
	add	$32,	%r9
	jmp	.aes_entry
	
	.align 16
.aes_loop:
	# middle of middle round
	movdqa  %xmm13,	%xmm4	# 4 : sb1u
	pshufb  %xmm2,	%xmm4   # 4 = sb1u
	pxor	(%r9),	%xmm4	# 4 = sb1u + k
	movdqa  %xmm12,	%xmm0	# 0 : sb1t
	pshufb  %xmm3,	%xmm0	# 0 = sb1t
	pxor	%xmm4,	%xmm0	# 0 = A
	movdqa  %xmm15,	%xmm4	# 4 : sb2u
	pshufb	%xmm2,	%xmm4	# 4 = sb2u
	movdqa	.k_mc_forward(%r11,%r10), %xmm1
	movdqa	%xmm14, %xmm2	# 2 : sb2t
	pshufb	%xmm3,  %xmm2	# 2 = sb2t
	pxor	%xmm4,  %xmm2	# 2 = 2A
	movdqa	%xmm0,  %xmm3	# 3 = A
	pshufb  %xmm1,  %xmm0	# 0 = B
	pxor	%xmm2,  %xmm0	# 0 = 2A+B
	pshufb	.k_mc_backward(%r11,%r10), %xmm3  # 3 = D
	pxor	%xmm0,	%xmm3	# 3 = 2A+B+D
	add	$16,	%r9	# next key
	pshufb  %xmm1,	%xmm0	# 0 = 2B+C
	pxor	%xmm3,	%xmm0	# 0 = 2A+3B+C+D
	add	$16, %r11	# next mc
	and	$48, %r11	# ... mod 4
	dec	%rax		# nr--

.aes_entry:
	# top of round
	movdqa  %xmm9, 	%xmm1	# 1 : i
	pandn	%xmm0, 	%xmm1	# 1 = i<<4
	psrld	$4,    	%xmm1   # 1 = i
	pand	%xmm9, 	%xmm0   # 0 = k
	movaps	%xmm11, %xmm2	# 2 : a/k
	pshufb  %xmm0,  %xmm2	# 2 = a/k
	pxor	%xmm1,	%xmm0	# 0 = j
	movaps  %xmm10,	%xmm3  	# 3 : 1/i
	pshufb  %xmm1, 	%xmm3  	# 3 = 1/i
	pxor	%xmm2, 	%xmm3  	# 3 = iak = 1/i + a/k
	movaps	%xmm10,	%xmm4  	# 4 : 1/j
	pshufb	%xmm0, 	%xmm4  	# 4 = 1/j
	pxor	%xmm2, 	%xmm4  	# 4 = jak = 1/j + a/k
	movaps  %xmm10,	%xmm2  	# 2 : 1/iak
	pshufb  %xmm3,	%xmm2  	# 2 = 1/iak
	pxor	%xmm0, 	%xmm2  	# 2 = io
	movaps  %xmm10, %xmm3   # 3 : 1/jak
	pshufb  %xmm4,  %xmm3   # 3 = 1/jak
	pxor	%xmm1,  %xmm3   # 3 = jo
	jnz	.aes_loop

	# middle of last round
	movdqa	.k_sbo(%r10), %xmm4	# 3 : sbou
	pshufb  %xmm2,  %xmm4   # 4 = sbou
	pxor	(%r9),  %xmm4   # 4 = sb1u + k
	movdqa	.k_sbo+16(%r10), %xmm0	# 0 : sbot
	pshufb  %xmm3,	%xmm0	# 0 = sb1t
	pxor	%xmm4,	%xmm0	# 0 = A
	pshufb	.k_sr(%r11,%r10), %xmm0
	ret

##
##  .aes_cleanup
##
##  Erases sensitive registers %xmm0-%xmm8
##
_aes_cleanup:
	movdqa	%xmm9,  %xmm0
	movdqa	%xmm9,  %xmm1
	movdqa	%xmm9,  %xmm2
	movdqa	%xmm9,  %xmm3
	movdqa	%xmm9,  %xmm4
	movdqa	%xmm9,  %xmm5
	movdqa	%xmm9,  %xmm6
	movdqa	%xmm9,  %xmm7
	movdqa  %xmm9, 	%xmm8
	ret
	
##
##  Decryption core
##
##  Same API as encryption core, except that it clobbers
##  %xmm5.  It's actually not clear that this is worthwhile.
##
_aes_decrypt_core:
	mov	%rdx,	%r9		# load key
	mov	(%rdx),	%rax
	mov	%rax,	%r11
	shl	$4,	%r11
	xor	$48,	%r11
	and	$48,	%r11
	movdqa	.k_dipt   (%r10), %xmm2 # iptlo
	movdqa	%xmm9,	%xmm1
	pandn	%xmm0,	%xmm1
	psrld	$4,	%xmm1
	pand	%xmm9,	%xmm0
	pshufb	%xmm0,	%xmm2
	movdqa	.k_dipt+16(%r10), %xmm0 # ipthi
	pshufb	%xmm1,	%xmm0
	pxor	16(%rdx),%xmm2
	pxor	%xmm2,	%xmm0
	add	$32,	%r9
	movdqa	.k_mc_forward+48(%r10), %xmm5
	jmp	.aes_dec_entry

	.align 16
.aes_dec_loop:
##
##  Inverse mix columns
##
	movdqa	(%r9),	%xmm0
	movdqa  .k_dsb9(%r10),%xmm4	# 4 : sb9u
	pshufb	%xmm2,	%xmm4		# 4 = sb9u
	pxor	%xmm0,	%xmm4
	movdqa  .k_dsb9+16(%r10),%xmm0	# 0 : sb9t
	pshufb	%xmm3,	%xmm0		# 0 = sb9t
	pxor	%xmm4,	%xmm0		# 0 = ch
	add	$16, %r9		# next round key

	pshufb	%xmm5,	%xmm0		# MC ch
	movdqa  .k_dsbd(%r10),%xmm4	# 4 : sbdu
	pshufb	%xmm2,	%xmm4		# 4 = sbdu
	pxor	%xmm0,	%xmm4		# 4 = ch
	movdqa  .k_dsbd+16(%r10),%xmm0	# 0 : sbdt
	pshufb	%xmm3,	%xmm0		# 0 = sbdt
	pxor	%xmm4,	%xmm0		# 0 = ch
	dec	%rax			# nr--
	
	pshufb	%xmm5,	%xmm0		# MC ch
	movdqa  .k_dsbb(%r10),%xmm4	# 4 : sbbu
	pshufb	%xmm2,	%xmm4		# 4 = sbbu
	pxor	%xmm0,	%xmm4		# 4 = ch
	movdqa  .k_dsbb+16(%r10),%xmm0	# 0 : sbbt
	pshufb	%xmm3,	%xmm0		# 0 = sbbt
	pxor	%xmm4,	%xmm0		# 0 = ch
	
	pshufb	%xmm5,	%xmm0		# MC ch
	movdqa  .k_dsbe(%r10),%xmm4	# 4 : sbeu
	pshufb	%xmm2,	%xmm4		# 4 = sbeu
	pxor	%xmm0,	%xmm4		# 4 = ch
	movdqa  .k_dsbe+16(%r10),%xmm0	# 0 : sbet
	pshufb	%xmm3,	%xmm0		# 0 = sbet
	pxor	%xmm4,	%xmm0		# 0 = ch

	palignr	$12,	%xmm5,	%xmm5
	
.aes_dec_entry:
	# top of round
	movdqa  %xmm9, 	%xmm1	# 1 : i
	pandn	%xmm0, 	%xmm1	# 1 = i<<4
	psrld	$4,    	%xmm1   # 1 = i
	pand	%xmm9, 	%xmm0   # 0 = k
	movaps	%xmm11, %xmm2	# 2 : a/k
	pshufb  %xmm0,  %xmm2	# 2 = a/k
	pxor	%xmm1,	%xmm0	# 0 = j
	movaps  %xmm10,	%xmm3  	# 3 : 1/i
	pshufb  %xmm1, 	%xmm3  	# 3 = 1/i
	pxor	%xmm2, 	%xmm3  	# 3 = iak = 1/i + a/k
	movaps	%xmm10,	%xmm4  	# 4 : 1/j
	pshufb	%xmm0, 	%xmm4  	# 4 = 1/j
	pxor	%xmm2, 	%xmm4  	# 4 = jak = 1/j + a/k
	movaps  %xmm10,	%xmm2  	# 2 : 1/iak
	pshufb  %xmm3,	%xmm2  	# 2 = 1/iak
	pxor	%xmm0, 	%xmm2  	# 2 = io
	movaps  %xmm10, %xmm3   # 3 : 1/jak
	pshufb  %xmm4,  %xmm3   # 3 = 1/jak
	pxor	%xmm1,  %xmm3   # 3 = jo
	jnz	.aes_dec_loop

	# middle of last round
	movdqa	.k_dsbo(%r10), %xmm4	# 3 : sbou
	pshufb  %xmm2,  %xmm4   # 4 = sbou
	pxor	(%r9),  %xmm4   # 4 = sb1u + k
	movdqa	.k_dsbo+16(%r10), %xmm0	# 0 : sbot
	pshufb  %xmm3,	%xmm0	# 0 = sb1t
	pxor	%xmm4,	%xmm0	# 0 = A
	pshufb	.k_sr(%r11,%r10), %xmm0
	ret

########################################################
##                                                    ##
##                     Constants                      ##
##                                                    ##
########################################################
	
	.align 16
.aes_consts:
	# s0F
	.k_s0F = .-.aes_consts
	.quad	0x0F0F0F0F0F0F0F0F
	.quad	0x0F0F0F0F0F0F0F0F

	# input transform (lo, hi)
	.k_ipt = .-.aes_consts
	.quad	0xC2B2E8985A2A7000
	.quad	0xCABAE09052227808
	.quad	0x4C01307D317C4D00
	.quad	0xCD80B1FCB0FDCC81

	# inv, inva
	.k_inv = .-.aes_consts
	.quad	0x0E05060F0D080180
	.quad	0x040703090A0B0C02
	.quad	0x01040A060F0B0780
	.quad	0x030D0E0C02050809

	# sb1u, sb1t
	.k_sb1 = .-.aes_consts
	.quad	0xB19BE18FCB503E00
	.quad	0xA5DF7A6E142AF544
	.quad	0x3618D415FAE22300
	.quad	0x3BF7CCC10D2ED9EF


	# sb2u, sb2t
	.k_sb2 = .-.aes_consts
	.quad	0xE27A93C60B712400
	.quad	0x5EB7E955BC982FCD
	.quad	0x69EB88400AE12900
	.quad	0xC2A163C8AB82234A

	# sbou, sbot
	.k_sbo = .-.aes_consts
	.quad	0xD0D26D176FBDC700
	.quad	0x15AABF7AC502A878
	.quad	0xCFE474A55FBB6A00
	.quad	0x8E1E90D1412B35FA

	# mc_forward
	.k_mc_forward = .-.aes_consts
	.quad	0x0407060500030201
	.quad	0x0C0F0E0D080B0A09
	.quad	0x080B0A0904070605
	.quad	0x000302010C0F0E0D
	.quad	0x0C0F0E0D080B0A09
	.quad	0x0407060500030201
	.quad	0x000302010C0F0E0D
	.quad	0x080B0A0904070605

	# mc_backward
	.k_mc_backward = .-.aes_consts
	.quad	0x0605040702010003
	.quad	0x0E0D0C0F0A09080B
	.quad	0x020100030E0D0C0F
	.quad	0x0A09080B06050407
	.quad	0x0E0D0C0F0A09080B
	.quad	0x0605040702010003
	.quad	0x0A09080B06050407
	.quad	0x020100030E0D0C0F

	# sr
	.k_sr = .-.aes_consts
	.quad	0x0706050403020100
	.quad	0x0F0E0D0C0B0A0908
	.quad	0x030E09040F0A0500
	.quad	0x0B06010C07020D08
	.quad	0x0F060D040B020900
	.quad	0x070E050C030A0108
	.quad	0x0B0E0104070A0D00
	.quad	0x0306090C0F020508

	# rcon
	.k_rcon = .-.aes_consts
	.quad	0x1F8391B9AF9DEEB6
	.quad	0x702A98084D7C7D81

	# s63: all equal to 0x63 transformed
	.k_s63 = .-.aes_consts
	.quad	0x5B5B5B5B5B5B5B5B
	.quad	0x5B5B5B5B5B5B5B5B

	# output transform
	.k_opt = .-.aes_consts
	.quad	0xFF9F4929D6B66000
	.quad	0xF7974121DEBE6808
	.quad	0x01EDBD5150BCEC00
	.quad	0xE10D5DB1B05C0CE0

	# deskew tables: inverts the sbox's "skew"
	.k_deskew = .-.aes_consts
	.quad	0x07E4A34047A4E300
	.quad	0x1DFEB95A5DBEF91A
	.quad	0x5F36B5DC83EA6900
	.quad	0x2841C2ABF49D1E77

##
##  Decryption stuff
##  Key schedule constants
##
	# decryption key schedule: x -> invskew x*9
	.k_dks_1 = .-.aes_consts
	.quad	0xB6116FC87ED9A700
	.quad	0x4AED933482255BFC
	.quad	0x4576516227143300
	.quad	0x8BB89FACE9DAFDCE

	# decryption key schedule: invskew x*9 -> invskew x*D
	.k_dks_2 = .-.aes_consts
	.quad	0x27438FEBCCA86400
	.quad	0x4622EE8AADC90561
	.quad	0x815C13CE4F92DD00
	.quad	0x73AEE13CBD602FF2
	
	# decryption key schedule: invskew x*D -> invskew x*B
	.k_dks_3 = .-.aes_consts
	.quad	0x03C4C50201C6C700
	.quad	0xF83F3EF9FA3D3CFB
	.quad	0xEE1921D638CFF700
	.quad	0xA5526A9D7384BC4B
	
	# decryption key schedule: invskew x*B -> invskew x*E + 0x63
	.k_dks_4 = .-.aes_consts
	.quad	0xE3C390B053732000
	.quad	0xA080D3F310306343
	.quad	0xA0CA214B036982E8
	.quad	0x2F45AEC48CE60D67

##
##  Decryption stuff
##  Round function constants
##
	# decryption input transform
	.k_dipt = .-.aes_consts
	.quad	0x0F505B040B545F00
	.quad	0x154A411E114E451A
	.quad	0x86E383E660056500
	.quad	0x12771772F491F194

	# decryption sbox output *9*u, *9*t
	.k_dsb9 = .-.aes_consts
	.quad	0x851C03539A86D600
	.quad	0xCAD51F504F994CC9
	.quad	0xC03B1789ECD74900
	.quad	0x725E2C9EB2FBA565

	# decryption sbox output *D*u, *D*t
	.k_dsbd = .-.aes_consts
	.quad	0x7D57CCDFE6B1A200
	.quad	0xF56E9B13882A4439
	.quad	0x3CE2FAF724C6CB00
	.quad	0x2931180D15DEEFD3

	# decryption sbox output *B*u, *B*t
	.k_dsbb = .-.aes_consts
	.quad	0xD022649296B44200
	.quad	0x602646F6B0F2D404
	.quad	0xC19498A6CD596700
	.quad	0xF3FF0C3E3255AA6B

	# decryption sbox output *E*u, *E*t
	.k_dsbe = .-.aes_consts
	.quad	0x46F2929626D4D000
	.quad	0x2242600464B4F6B0
	.quad	0x0C55A6CDFFAAC100
	.quad	0x9467F36B98593E32
	
	# decryption sbox final output
	.k_dsbo = .-.aes_consts
	.quad	0x1387EA537EF94000
	.quad	0xC7AA6DB9D4943E2D
	.quad	0x12D7560F93441D00
	.quad	0xCA4B8159D8C58E9C

	.k_ctr_mask = .-.aes_consts
	.quad	0x0000000000000000
	.quad	0xFFFFFFFFFFFFFFFF

	.k_ctr_one = .-.aes_consts
	.quad	0x0000000000000001
	.quad	0x0000000000000000
	
	# ocb mode
	.k_ocb_mask = .-.aes_consts
	.quad	0x0100000001000000
	.quad	0x8680000001000000

	# padding
	.k_reverse_id = .-.aes_consts
	.quad	0x08090A0B0C0D0E0F
	.quad	0x0001020304050607

########################################################
##                                                    ##
##                  AES key schedule                  ##
##                                                    ##
########################################################
	
aes_schedule_core:
	# rdi = key
	# rsi = size in bits
	# rdx = buffer
	# rcx = direction.  0=encrypt, 1=decrypt

	call	_aes_preheat		# load the tables
	movdqa	.k_rcon(%r10), %xmm8	# load rcon
	movdqu	(%rdi),	%xmm0		# load key (unaligned)

	# input transform
	movdqu	%xmm0,	%xmm3
	lea	.k_ipt(%r10), %r11
	call	.aes_schedule_transform
	movdqu	%xmm0,	%xmm7

	test	%rcx,	%rcx
	jnz	.aes_schedule_am_decrypting

	# encrypting, output zeroth round key after transform
	movdqa	%xmm0,	(%rdx)
	jmp	.aes_schedule_go
	
.aes_schedule_am_decrypting:
	# decrypting, output zeroth round key after shiftrows
	pshufb  .k_sr(%r8,%r10),%xmm3
	movdqa	%xmm3,	(%rdx)
	xor	$48, 	%r8
	
.aes_schedule_go:
	cmp	$192,	%rsi
	je	.aes_schedule_192
	cmp	$256,	%rsi
	je	.aes_schedule_256
	# 128: fall though

##
##  .aes_schedule_128
##
##  128-bit specific part of key schedule.
##
##  This schedule is really simple, because all its parts
##  are accomplished by the subroutines.
##
.aes_schedule_128:
	mov	$10, %rsi
	
.aes_schedule_128_L:
	call 	.aes_schedule_round
	dec	%rsi
	jz 	.aes_schedule_mangle_last
	call	.aes_schedule_mangle	# write output
	jmp 	.aes_schedule_128_L

##
##  .aes_schedule_192
##
##  192-bit specific part of key schedule.
##
##  The main body of this schedule is the same as the 128-bit
##  schedule, but with more smearing.  The long, high side is
##  stored in %xmm7 as before, and the short, low side is in
##  the high bits of %xmm6.
##
##  This schedule is somewhat nastier, however, because each
##  round produces 192 bits of key material, or 1.5 round keys.
##  Therefore, on each cycle we do 2 rounds and produce 3 round
##  keys.
##
.aes_schedule_192:
	movdqu	8(%rdi),%xmm0		# load key part 2 (very unaligned)
	call	.aes_schedule_transform	# input transform
	movdqa	%xmm0,	%xmm6		# save short part
	pxor	%xmm4,	%xmm4		# clear 4
	movhlps	%xmm4,	%xmm6		# clobber low side with zeros
	mov	$4,	%rsi

.aes_schedule_192_L:
	call	.aes_schedule_round
	palignr	$8,%xmm6,%xmm0	
	call	.aes_schedule_mangle	# save key n
	call	.aes_schedule_192_smear
	call	.aes_schedule_mangle	# save key n+1
	call	.aes_schedule_round
	dec	%rsi
	jz 	.aes_schedule_mangle_last
	call	.aes_schedule_mangle	# save key n+2
	call	.aes_schedule_192_smear
	jmp	.aes_schedule_192_L

##
##  .aes_schedule_192_smear
##
##  Smear the short, low side in the 192-bit key schedule.
##
##  Inputs:
##    %xmm7: high side, b  a  x  y
##    %xmm6:  low side, d  c  0  0
##    %xmm13: 0
##
##  Outputs:
##    %xmm6: b+c+d  b+c  0  0
##    %xmm0: b+c+d  b+c  b  a
##
.aes_schedule_192_smear:
	pshufd	$0x80,	%xmm6,	%xmm0	# d c 0 0 -> c 0 0 0
	pxor	%xmm0,	%xmm6		# -> c+d c 0 0
	pshufd	$0xFE,	%xmm7,	%xmm0	# b a _ _ -> b b b a
	pxor	%xmm0,	%xmm6		# -> b+c+d b+c b a
	movdqa	%xmm6,	%xmm0
	pxor	%xmm1,	%xmm1
	movhlps	%xmm1,	%xmm6		# clobber low side with zeros
	ret

##
##  .aes_schedule_256
##
##  256-bit specific part of key schedule.
##
##  The structure here is very similar to the 128-bit
##  schedule, but with an additional "low side" in
##  %xmm6.  The low side's rounds are the same as the
##  high side's, except no rcon and no rotation.
##
.aes_schedule_256:
	movdqu	16(%rdi),%xmm0		# load key part 2 (unaligned)
	call	.aes_schedule_transform	# input transform
	mov	$7, %rsi
	
.aes_schedule_256_L:
	call	.aes_schedule_mangle	# output low result
	movdqa	%xmm0,	%xmm6		# save cur_lo in xmm6

	# high round
	call	.aes_schedule_round
	dec	%rsi
	jz 	.aes_schedule_mangle_last
	call	.aes_schedule_mangle	

	# low round. swap xmm7 and xmm6
	pshufd	$0xFF,	%xmm0,	%xmm0
	movdqa	%xmm7,	%xmm5
	movdqa	%xmm6,	%xmm7
	call	.aes_schedule_low_round
	movdqa	%xmm5,	%xmm7
	
	jmp	.aes_schedule_256_L

##
##  .aes_schedule_round
##
##  Runs one main round of the key schedule on %xmm0, %xmm7
##
##  Specifically, runs subbytes on the high dword of %xmm0
##  then rotates it by one byte and xors into the low dword of
##  %xmm7.
##
##  Adds rcon from low byte of %xmm8, then rotates %xmm8 for
##  next rcon.
##
##  Smears the dwords of %xmm7 by xoring the low into the
##  second low, result into third, result into highest.
##
##  Returns results in %xmm7 = %xmm0.
##  Clobbers %xmm1-%xmm4, %r11.
##
.aes_schedule_round:
	# extract rcon from xmm8
	pxor	%xmm1,	%xmm1
	palignr	$15,	%xmm8,	%xmm1
	palignr	$15,	%xmm8,	%xmm8
	pxor	%xmm1,	%xmm7

	# rotate
	pshufd	$0xFF,	%xmm0,	%xmm0
	palignr	$1,	%xmm0,	%xmm0
	
	# fall through...
	
	# low round: same as high round, but no rotation and no rcon.
.aes_schedule_low_round:
	# smear xmm7
	movdqa	%xmm7,	%xmm1
	pslldq	$4,	%xmm7
	pxor	%xmm1,	%xmm7
	movdqa	%xmm7,	%xmm1
	pslldq	$8,	%xmm7
	pxor	%xmm1,	%xmm7
	pxor	.k_s63(%r10), %xmm7

	# subbytes
	movdqa  %xmm9, 	%xmm1
	pandn	%xmm0, 	%xmm1
	psrld	$4,    	%xmm1		# 1 = i
	pand	%xmm9, 	%xmm0		# 0 = k
	movaps	%xmm11, %xmm2		# 2 : a/k
	pshufb  %xmm0,  %xmm2		# 2 = a/k
	pxor	%xmm1,	%xmm0		# 0 = j
	movaps  %xmm10,	%xmm3		# 3 : 1/i
	pshufb  %xmm1, 	%xmm3		# 3 = 1/i
	pxor	%xmm2, 	%xmm3		# 3 = iak = 1/i + a/k
	movaps	%xmm10,	%xmm4		# 4 : 1/j
	pshufb	%xmm0, 	%xmm4		# 4 = 1/j
	pxor	%xmm2, 	%xmm4		# 4 = jak = 1/j + a/k
	movaps  %xmm10,	%xmm2		# 2 : 1/iak
	pshufb  %xmm3,	%xmm2		# 2 = 1/iak
	pxor	%xmm0, 	%xmm2		# 2 = io
	movaps  %xmm10, %xmm3		# 3 : 1/jak
	pshufb  %xmm4,  %xmm3		# 3 = 1/jak
	pxor	%xmm1,  %xmm3		# 3 = jo
	movdqa	.k_sb1(%r10), %xmm4	# 4 : sbou
	pshufb  %xmm2,  %xmm4		# 4 = sbou
	movdqa	.k_sb1+16(%r10), %xmm0	# 0 : sbot
	pshufb  %xmm3,	%xmm0		# 0 = sb1t
	pxor	%xmm4, 	%xmm0		# 0 = sbox output

	# add in smeared stuff
	pxor	%xmm7,	%xmm0	
	movdqa	%xmm0,	%xmm7
	ret

##
##  .aes_schedule_transform
##
##  Linear-transform %xmm0 according to tables at (%r11)
##
##  Requires that %xmm9 = 0x0F0F... as in preheat
##  Output in %xmm0
##  Clobbers %xmm1, %xmm2
##
.aes_schedule_transform:
	movdqa	%xmm9,	%xmm1
	pandn	%xmm0,	%xmm1
	psrld	$4,	%xmm1
	pand	%xmm9,	%xmm0
	movaps	(%r11), %xmm2 	# lo
	pshufb	%xmm0,	%xmm2
	movaps	16(%r11), %xmm0 # hi
	pshufb	%xmm1,	%xmm0
	pxor	%xmm2,	%xmm0
	ret

##
##  .aes_schedule_mangle
##
##  Mangle xmm0 from (basis-transformed) standard version
##  to our version.
##
##  On encrypt,
##    xor with 0x63
##    multiply by circulant 0,1,1,1
##    apply shiftrows transform
##
##  On decrypt,
##    xor with 0x63
##    multiply by "inverse mixcolumns" circulant E,B,D,9
##    deskew
##    apply shiftrows transform
##
##
##  Writes out to (%rdx), and increments or decrements it
##  Keeps track of round number mod 4 in %r8
##  Preserves xmm0
##  Clobbers xmm1-xmm5
##
.aes_schedule_mangle:
	movdqa	%xmm0,	%xmm4	# save xmm0 for later
	movdqa	.k_mc_forward(%r10),%xmm5
	test	%rcx, 	%rcx
	jnz	.aes_schedule_mangle_dec

	# encrypting
	add	$16,	%rdx
	pxor	.k_s63(%r10),%xmm4
	pshufb	%xmm5,	%xmm4
	movdqa	%xmm4,	%xmm3
	pshufb	%xmm5,	%xmm4
	pxor	%xmm4,	%xmm3
	pshufb	%xmm5,	%xmm4
	pxor	%xmm4,	%xmm3

	jmp	.aes_schedule_mangle_both
	
.aes_schedule_mangle_dec:
	lea	.k_dks_1(%r10), %r11	# first table: *9
	call 	.aes_schedule_transform
	movdqa	%xmm0,	%xmm3
	pshufb	%xmm5,	%xmm3
	
	add	$32, 	%r11		# next table:  *B
	call 	.aes_schedule_transform
	pxor	%xmm0,	%xmm3
	pshufb	%xmm5,	%xmm3
	
	add	$32, 	%r11		# next table:  *D
	call 	.aes_schedule_transform
	pxor	%xmm0,	%xmm3
	pshufb	%xmm5,	%xmm3
	
	add	$32, 	%r11		# next table:  *E
	call 	.aes_schedule_transform
	pxor	%xmm0,	%xmm3
	pshufb	%xmm5,	%xmm3

	movdqa	%xmm4,	%xmm0		# restore %xmm0
	add	$-16,	%rdx

.aes_schedule_mangle_both:
	pshufb	.k_sr(%r8,%r10),%xmm3
	add	$-16,	%r8
	and	$48,	%r8
	movdqa	%xmm3,	(%rdx)
	ret
	
##
##  .aes_schedule_mangle_last
##
##  Mangler for last round of key schedule
##  Mangles %xmm0
##    when encrypting, outputs out(%xmm0) ^ 63
##    when decrypting, outputs unskew(%xmm0)
##
##  Always called right before return... jumps to cleanup and exits
##
.aes_schedule_mangle_last:
	# schedule last round key from xmm0
	lea	.k_deskew(%r10),%r11	# prepare to deskew
	test	%rcx, 	%rcx
	jnz	.aes_schedule_mangle_last_dec

	# encrypting
	pshufb	.k_sr(%r8,%r10),%xmm0	# output permute
	lea	.k_opt(%r10),	%r11	# prepare to output transform
	add	$32,	%rdx

.aes_schedule_mangle_last_dec:
	add	$-16,	%rdx
	pxor	.k_s63(%r10),	%xmm0
	call	.aes_schedule_transform # output transform
	movdqa	%xmm0,	(%rdx)		# save last key
	jmp 	_aes_cleanup
	
#########################################################
##                                                     ##
##                      Padding                        ##
##                                                     ##
#########################################################

##
## Check padding in %xmm0.
## Return unpadded size in %rcx.
##
## Padding mode in %rax, requires %r10 preloaded
## Clobber %r8
##
##   AES_PADDING_NULL        = 0
##   AES_PADDING_DES         = 1
##   AES_PADDING_PKCS5       = 2
##
	.align 16
_check_padding:
	cmp	$0,	%rax
	jz	.cp_null
	cmp	$1,	%rax
	jz	.cp_des
	cmp	$2,	%rax
	jz	.cp_pkcs5
	mov	$-1,	%rcx
	ret

.cp_null:
	pxor	%xmm1,	%xmm1
	pcmpeqb	%xmm0,	%xmm1
	pmovmskb %xmm1,	%rcx		# r8 = mask of zero bytes
	not	%cx			# ... inverted
	bsr	%cx,	%cx		# highest bit set (from 0)
	jnz	.cp_notnull_16
	mov	$15,	%cx
	
.cp_notnull_16:
	inc	%rcx
	ret
	
.cp_des:
	pxor	%xmm2,	%xmm2		# v2 = 0
	mov	$0x80,	%rcx
	movd	%rcx,	%xmm1
	pshufb	%xmm2,	%xmm1		# v1 = all 0x80
	pcmpeqb	%xmm0,	%xmm1		# v1 = match 0x80
	pcmpeqb	%xmm0,	%xmm2		# v2 = match 0
	pmovmskb %xmm2,	%rcx		# rcx = mask of 0s
	not	%cx
	bsr	%cx,	%cx		# find furthest non-0
	jz	.cp_fail
	pmovmskb %xmm1,	%r8		# r8 = mask of 0x80s
	bsr	%r8,	%r8		# find furthest 0x80
	jz	.cp_fail
	cmp	%r8,	%rcx		# same?
	jnz	.cp_fail
	ret
	
	
.cp_pkcs5:
	movdqa	.k_reverse_id(%r10), %xmm3	# v3 = (...,3,2,1,0)
	movdqa	%xmm0,	%xmm1
	pshufb	.k_s0F(%r10),	%xmm1	# v1 = repeat first byte
	movdqa	%xmm1,	%xmm2
	movd	%xmm1,	%r8		# r8 = repeat first byte
	pcmpgtb	%xmm3,	%xmm2		# v2 = (0*(16-n), FF*n)
	pcmpeqb	%xmm0,	%xmm1		# check if == last byte
	pandn	%xmm2,	%xmm1		# anything amiss?
	pmovmskb %xmm1,	%rcx
	test	%rcx,	%rcx
	jne	.cp_fail

	and	$255,	%r8
	cmp	$16,	%r8
	ja	.cp_fail
	mov	$16,	%rcx
	sub	%r8,	%rcx
	ret
	
.cp_fail:
	mov 	$-1, %rcx
	ret


##
## Load %rcx bytes from (%rdi) into %xmm0.
## Require 0 <= %rcx <= 16.
## The rest of the bytes will be masked off to 0.
## Mask output in %xmm1.
##
## Padding mode in %rax, chaining block in %xmm0 for
## ciphertext stealing.  Padding mode must be one of:
##	
##   AES_PADDING_NULL        = 0
##   AES_PADDING_DES         = 1
##   AES_PADDING_PKCS5       = 2
##   AES_CIPHERTEXT_STEALING = 3
##
## Anything else, and you get NULL.
##
## Require %r10 preheated.
##
## Clobbers %r8, %xmm0-%xmm4.
##

	.align 16
_unaligned_partial_load:
	movdqa	%xmm0,	%xmm4
	pxor	%xmm1,	%xmm1		# v1 = 0
	pxor	%xmm2,	%xmm2		# v2 = 0
	test	%rcx,	%rcx
	jz	.upl_none
	
	mov 	%rdi,	%r8
	and 	$15,	%r8		# r8 = offset
	movd	%r8,	%xmm1		# v1 = offset
	movdqa	.k_sr(%r10), %xmm3	# v3 = (0,1,2,...)
	add 	%rcx, 	%r8		# r8 = end-offset mod 16
	cmp 	$16, 	%r8
	jle	.upl_within

.upl_across:				# across a 16-byte boundary
	movdqu	(%rdi), %xmm0		# no danger of segfault
	jmp	.upl_loaded

.upl_within:				# within a 16-byte aligned load
	mov	%rdi,	%r8
	and	$-16,	%r8		# r8 = offset aligned
	movdqa	(%r8),	%xmm0		# load aligned
	pshufb	%xmm2,	%xmm1		# v1 = (off,off,...)
	paddb	%xmm3,	%xmm1		# v1 = (off+0,off+1,...)
	pshufb	%xmm1,	%xmm0		# v0 offset by off

.upl_loaded:
	movd	%rcx,	%xmm1		# v1 = n
	pshufb	%xmm2,	%xmm1		# v1 = (n,n,n,n,...)
	pcmpgtb	%xmm3,	%xmm1		# v1 = (0xFF * n, 0 * (16-n))

.upl_none:
	pand	%xmm1,	%xmm0		# mask off bytes
	
	cmp	$1,	%rax
	jz	.upl_padding_des
	cmp	$2,	%rax
	jz	.upl_padding_pkcs5
	cmp	$3,	%rax
	jz	.upl_padding_cts

.upl_padding_null:
	ret

.upl_padding_des:
	mov	$0x80,	%r8
	movd	%r8,	%xmm3	# xmm3 gets 0x80
	pcmpeqb	%xmm1,	%xmm2	# invert into 2
	pslldq	$1,	%xmm2	# shift off
	por	%xmm1,	%xmm2	# 0 in both -> 0, else FF
	pshufb	%xmm2,	%xmm3	# extract 0x80 there
	por	%xmm3,	%xmm0	# add to data
	ret

.upl_padding_pkcs5:
	mov	$16,	%r8
	sub	%rcx,	%r8	# r8 = pad
	movd	%r8,	%xmm3
	pshufb	%xmm2,	%xmm3	# xmm3: pad, pad, pad...
	pandn	%xmm3,	%xmm1	# clear pads
	por	%xmm1,	%xmm0	# add to data
	ret

.upl_padding_cts:
	pandn	%xmm4,	%xmm1
	por	%xmm1,	%xmm0
	ret
##
## Unaligned partial store.
##
## Store %rcx bytes of %xmm0 into (%rsi).
##
## Clobbers %rdi, %xmm1-%xmm5.
##
	
	.align 16
_unaligned_partial_store:
	test	%rcx,	%rcx
	jz	.ups_none
	movdqa	%xmm0,	%xmm5		# r5 = copy of input
	mov 	%rsi,	%rdi
	and 	$15,	%rdi		# r8 = offset mod 16
	movdqa	.k_sr(%r10), %xmm3	# v3 = (0,1,2,...)
	pxor	%xmm2,	%xmm2		# v2 = 0
	movd	%rcx,	%xmm4		# v4 = n
	pshufb	%xmm2,	%xmm4		# v4 = (n,n,n,n...)
	pcmpgtb	%xmm3,	%xmm4		# v4 = mask = (0xFF * n, 0 * (16-n))
	movd	%rdi,	%xmm1		# v1 = offset

	add 	%rcx, 	%rdi		# r8 = end-offset mod 16
	cmp 	$16, 	%rdi
	mov	%rsi,	%rdi
	ja	.ups_across

.ups_within:				# within a 16-byte aligned store
	and	$-16,	%rdi		# r8 = offset aligned
	pshufb	%xmm2,	%xmm1		# v1 = (off,off,off...)
	psubb	%xmm1,	%xmm3		# v1 = (0-off, 1-off, ...)
	pshufb	%xmm3,	%xmm5		# shift input by offset
	pshufb	%xmm3,	%xmm4		# shift mask by offset
.ups_across:
	maskmovdqu	%xmm4,	%xmm5
.ups_none:
	ret

