#!/usr/bin/env perl

# simple benchmarking script using ./test

my @key_sizes = (128, 192, 256);
my @pt_sizes = (16, 32, 128, 576, 1024, 4096);
my @prfmodes = qw/ctr ofb/;
my @prpmodes = qw/cfb ecb cbc/;
my $niters = 100;
my $trials = 100;

my $header_length = 17 + 8*$#pt_sizes;

print "bench.pl: median of $trials trials, $niters iterations each\n\n";

sub print_centered {
	my $p = shift;
	my $h = $header_length - length $p;
	my $l = int($h/2) - 1;
	my $l2 = $h - $l - 2;
	print "="x$l, " $p ", "="x$l2, "\n";
}

sub size_header {
	print "size     " ;
	printf("%8d", $_) foreach (@pt_sizes);
	print "\n"; #, "="x$header_length, "\n";
}

sub trial {
	my $cmd = shift;
	my $out = `$cmd`;
	$out =~ s/.*?(\d+\.\d+) cp[bi]\n/$1/;
	return $out;
}

foreach $ks (@key_sizes) {
	print_centered("$ks-bit keys");
	
	size_header();
	foreach my $mode (@prfmodes, @prpmodes) {
		print "$mode      ";
		foreach my $ps(@pt_sizes) {
			printf "%8.2f",
			trial("./test -b $trials -n $niters -k $ks -m $mode -s $ps");
			sleep 0.25;
		}
		print "\n";
	}
	foreach my $mode (@prpmodes) {
		print "$mode (dec)";
		foreach my $ps(@pt_sizes) {
			printf "%8.2f",
			trial("./test -b $trials -n $niters -k $ks -m $mode -s $ps -d");
		}
		print "\n";
	}
	printf "sch       %7.2f\nsch (dec) %7.2f\n\n", 
		trial("./test -B $trials -n $niters -k $ks"),
		trial("./test -B $trials -n $niters -k $ks -d");
}
